@@ -1,6 +1,6 @@
 /* main.c
  *
- * Copyright (C) 2014 Christian Hergert <christian@hergert.me>
+ * Copyright © 2014 Christian Hergert <christian@hergert.me>
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -21,8 +21,13 @@
 #include <ide.h>
 #include <gtksourceview/gtksource.h>
 
+#include "application/ide-application-private.h"
+#include "plugins/gnome-builder-plugins.h"
+
 #include "bug-buddy.h"
 
+static IdeApplicationMode early_mode;
+
 static gboolean
 verbose_cb (const gchar  *option_name,
             const gchar  *value,
@@ -34,12 +39,14 @@ verbose_cb (const gchar  *option_name,
 }
 
 static void
-early_verbose_check (gint    *argc,
-                     gchar ***argv)
+early_params_check (gint    *argc,
+                    gchar ***argv)
 {
-  GOptionContext *context;
-  static const GOptionEntry entries[] = {
+  g_autofree gchar *type = NULL;
+  g_autoptr(GOptionContext) context = NULL;
+  GOptionEntry entries[] = {
     { "verbose", 'v', G_OPTION_FLAG_NO_ARG, G_OPTION_ARG_CALLBACK, verbose_cb },
+    { "type", 0, 0, G_OPTION_ARG_STRING, &type },
     { NULL }
   };
 
@@ -48,7 +55,11 @@ early_verbose_check (gint    *argc,
   g_option_context_set_help_enabled (context, FALSE);
   g_option_context_add_main_entries (context, entries, NULL);
   g_option_context_parse (context, argc, argv, NULL);
-  g_option_context_free (context);
+
+  if (g_strcmp0 (type, "worker") == 0)
+    early_mode = IDE_APPLICATION_MODE_WORKER;
+  else if (g_strcmp0 (type, "cli") == 0)
+    early_mode = IDE_APPLICATION_MODE_TOOL;
 }
 
 static void
@@ -92,11 +103,21 @@ main (int   argc,
   IdeApplication *app;
   int ret;
 
+  /* Setup our gdb fork()/exec() helper */
   bug_buddy_init ();
 
+  /* Early init of logging so that we get messages in a consistent
+   * format. If we deferred this to GApplication, we'd get them in
+   * multiple formats.
+   */
   ide_log_init (TRUE, NULL);
-  early_verbose_check (&argc, &argv);
 
+  /* Extract options like -vvvv and --type=worker only */
+  early_params_check (&argc, &argv);
+
+  /* We might need to prime SSL environment and other bits before
+   * the application has had a chance to setup caches/etc.
+   */
   early_ssl_check ();
 
   g_message ("Initializing with Gtk+ version %d.%d.%d.",
@@ -119,10 +140,26 @@ main (int   argc,
    */
   g_type_ensure (G_TYPE_ZLIB_DECOMPRESSOR);
 
+  /* Setup the application instance */
   app = ide_application_new ();
+  _ide_application_set_mode (app, early_mode);
+
+  /* Ensure that our static plugins init routine is called.
+   * This is necessary to ensure that -Wl,--as-needed does not
+   * drop our link to this shared library.
+   */
+  gnome_builder_plugins_init ();
+
+  /* Block until the application exits */
   ret = g_application_run (G_APPLICATION (app), argc, argv);
+
+  /* Force disposal of the application (to help catch cleanup
+   * issues at shutdown) and then (hopefully) finalize the app.
+   */
+  g_object_run_dispose (G_OBJECT (app));
   g_clear_object (&app);
 
+  /* Cleanup logging and flush anything that still needs it */
   ide_log_shutdown ();
 
   return ret;
